use std::{fmt::Debug, sync::Arc, time::Duration};

use domain::entities::account::Account;
use lazy_static::lazy_static;
use library::config;
use moka::{
    future::{Cache, CacheBuilder},
    notification::{ListenerFuture, RemovalCause},
    policy::EvictionPolicy,
};
use serde::{de::DeserializeOwned, Deserialize, Serialize};

use crate::cache::redis_cache::REDIS_CACHE;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct CacheAccount {
    pub account: Arc<Account>,
    pub token: Arc<String>,
}

pub struct ServerCache<S, D> {
    pub cache: Cache<S, D>,
    pub cache_type: Arc<String>,
}

type CacheEvictionListener<D> = Box<dyn Fn(Arc<String>, D, RemovalCause) -> ListenerFuture + Send + Sync + 'static>;

impl<D> ServerCache<String, D>
where
    D: Debug + Send + Sync + Clone + Serialize + DeserializeOwned + 'static,
{
    fn default(cache_type: &str) -> Self {
        let cache_type = Arc::new(cache_type.to_string());
        let cache_type_raw = cache_type.clone();
        let cache_eviction_listener: CacheEvictionListener<D> =
            Box::new(move |key: Arc<String>, value: D, cause: RemovalCause| {
                Box::pin({
                    let catch_type: Arc<String> = cache_type_raw.clone();
                    async move {
                        tracing::info!(
                            "内存缓存数据移除 key: {:?}, value: {:?}, cause: {:?}",
                            key,
                            value,
                            cause
                        );
                        // 如果缓存数据不是过期或被替换，则从redis中同步删除缓存数据
                        if cause != RemovalCause::Expired && cause != RemovalCause::Replaced {
                            RedisCache {
                                cache_type: catch_type,
                                cache_key: key.to_string(),
                                ..Default::default()
                            }
                            .remove()
                            .await;
                        }
                    }
                })
            });
        let cache = CacheBuilder::new(20480)
            .name("login_cache")
            .eviction_policy(EvictionPolicy::lru())
            .time_to_live(Duration::from_secs(config!().jwt.expires as u64))
            .async_eviction_listener(cache_eviction_listener)
            .build();
        ServerCache {
            cache,
            cache_type,
        }
    }

    /// 初始化缓存，从redis中读取缓存数据
    pub async fn init(&self) {
        let cache_datas = RedisCache::get(self.cache_type.clone()).await;
        for element in cache_datas {
            let cache_key = element.cache_key.clone();
            let data_json = element.data;
            let cache_data = serde_json::from_str::<D>(&data_json).unwrap();
            self.add_cache(cache_key, cache_data).await;
        }
    }

    /// 添加内存缓存
    async fn add_cache(&self, key: String, value: D) {
        self.cache.insert(key, value).await;
    }

    /// 添加缓存，包括内存缓存和redis缓存
    pub async fn insert(&self, key: String, value: D) {
        let cache_data = RedisCache {
            cache_type: self.cache_type.clone(),
            data: serde_json::to_string(&value).unwrap(),
            cache_key: key.to_string(),
            ..Default::default()
        };
        cache_data.insert().await;
        self.cache.insert(key, value).await;
    }

    /// 从内存获取缓存
    pub async fn get(&self, key: &str) -> Option<D> {
        self.cache.get(key).await
    }

    /// 从内存移除缓存
    pub async fn remove(&self, key: &str) {
        let cache_data = RedisCache {
            cache_type: self.cache_type.clone(),
            cache_key: key.to_string(),
            ..Default::default()
        };
        cache_data.remove().await;
        self.cache.remove(key).await;
    }
}

/// Redis缓存封装
#[derive(Debug, Clone, Default)]
struct RedisCache {
    pub cache_type: Arc<String>,
    pub cache_key: String,
    pub data: String,
}

impl RedisCache {
    /// 向redis缓存Hash中插入数据
    pub async fn insert(&self) {
        match REDIS_CACHE
            .hset_ex(
                &self.cache_type,
                &self.cache_key,
                &self.data,
                config!().jwt.expires,
            )
            .await
        {
            Ok(_) => {}
            Err(err) => {
                tracing::error!("向缓存Hash插入数据失败: {:?}", err);
            }
        }
    }

    /// 从redis缓存Hash中获取数据
    pub async fn get(cache_type: Arc<String>) -> Vec<RedisCache> {
        let mut result = Vec::new();
        match REDIS_CACHE.hgetall(&cache_type).await {
            Ok(cache_datas) => {
                for (key, value) in cache_datas {
                    let cache_data = RedisCache {
                        cache_type: cache_type.clone(),
                        cache_key: key,
                        data: value,
                    };
                    result.push(cache_data);
                }
            }
            Err(err) => {
                tracing::error!("从缓存Hash获取数据失败: {:?}", err);
            }
        }

        result
    }

    /// 从redis缓存Hash中删除数据
    pub async fn remove(&self) {
        match REDIS_CACHE.hdel(&self.cache_type, &self.cache_key).await {
            Ok(_) => {
                tracing::info!("从缓存Hash删除数据成功");
            }
            Err(err) => {
                tracing::error!("从缓存Hash删除数据失败: {:?}", err);
            }
        }
    }
}

pub async fn init_cache() {
    LOGIN_ACCOUNT_CACHE.init().await;
}

lazy_static! {
    pub static ref LOGIN_ACCOUNT_CACHE: ServerCache<String, Arc<CacheAccount>> = ServerCache::default("CACHE_ACCOUNT");
}
