use std::sync::Arc;

use chrono::Utc;
use domain::{
    dto::account::AuthenticateWithPassword,
    entities::account::{Account, Role},
    vo::account::LoginAccount,
};
use i18n::{
    message,
    message_ids::MessageId,
};
use library::{
    cache::inner_cache::{CacheAccount, LOGIN_ACCOUNT_CACHE}, context::Context, db, model::response::{ResErr, ResResult}, token::{generate_refresh_token, generate_token}
};

/// 登录, 使用账号和密码
pub async fn authenticate_with_password(
    context: Context,
    req: AuthenticateWithPassword,
) -> ResResult<LoginAccount> {
    let account =
        Account::find_with_password(req.username.unwrap(), req.password.unwrap(), db!()).await?;
    if account.is_none() {
        tracing::info!("登录用户失败，用户查询为空");
        return Err(ResErr::params(message!(
            context.get_lang_tag(),
            MessageId::IncorrectUsernameOrPassword
        )));
    }
    let account = account.unwrap();
    if let Some(disable_time) = account.disable_time {
        if disable_time > Utc::now() {
            tracing::error!("账户已禁用");
            return Err(ResErr::auth(message!(
                context.get_lang_tag(),
                MessageId::AccountDisabled
            )));
        }
    }
    if !account.role.is_admin() {
        tracing::error!("账户不是管理员，无权限");
        return Err(ResErr::perm(message!(
            context.get_lang_tag(),
            MessageId::AccountNoPermission
        )));
    }

    let token = generate_token(&account.id);
    let refresh_token = generate_refresh_token(&account.id);

    LOGIN_ACCOUNT_CACHE
        .insert(
            account.id.to_owned(),
            Arc::new(CacheAccount {
                account: Arc::new(account.to_owned()),
                token: Arc::new(token.to_owned()),
            }),
        )
        .await;

    let login_account = LoginAccount {
        username: account.username,
        display_name: account.display_name,
        avatar_url: account.avatar_url,
        metadata: account.metadata,
        wallet: account.wallet,
        email: account.email,
        token,
        refresh_token,
    };
    Ok(login_account)
}

/// 仅仅用于管理员维护，不对外暴露接口
pub async fn add_account() -> ResResult<()> {
    let mut transaction = db!().begin().await?;
    let account = match  Account::add_account(
        &Account {
            username: "admin".to_string(),
            password: Some("123456".as_bytes().to_vec()),
            role: Role::Admin,
            ..Default::default()
        },
        &mut transaction,
    )
    .await {
        Ok(account) => {
            transaction.commit().await?;
            account
        },
        Err(err) => {
            transaction.rollback().await?;
            tracing::error!(error = ?err, "添加用户失败");
            return Err(ResErr::service("添加用户失败"));
        }
    };

    if account.is_none() {
        tracing::error!("添加用户失败");
        return Err(ResErr::service("添加用户失败"));
    }
    Ok(())
}
